#include "file_cache.h"

const int FileCache::CACHE_DIR_LENGTH = 2;
const int FileCache::CACHE_DIR_LEVELS = 1;
const std::string FileCache::CACHE_LOCK_SUFFIX = ".lock";
const std::string FileCache::CACHE_META_SUFFIX = ".meta";

FileCache::FileCache(std::string cache_path,
    std::string cache_job_dir_path,
    std::string cache_link_path,
    std::string id,
    uid_t job_uid,
    gid_t job_gid) throw(FileCacheException) {
  
  // make a vector of one item and call _init
  struct CacheParameters cache_info;
  cache_info.cache_path = cache_path;
  cache_info.cache_job_dir_path = cache_job_dir_path;
  cache_info.cache_link_path = cache_link_path;
  
  std::vector<struct CacheParameters> caches;
  caches.push_back(cache_info);
  
  _init(caches, id, job_uid, job_gid);
}

FileCache::FileCache(std::vector<struct CacheParameters> caches,
    std::string id,
    uid_t job_uid,
    gid_t job_gid) throw(FileCacheException) {
  
  _init(caches, id, job_uid, job_gid);
}

void FileCache::_init(std::vector<struct CacheParameters> caches,
    std::string id,
    uid_t job_uid,
    gid_t job_gid) throw(FileCacheException) {
  
  _id = id;
  _uid = job_uid;
  _gid = job_gid;
  
  // for each cache
  for (int i = 0; i < (int)caches.size(); i++) {
    struct CacheParameters cache_params = caches.at(i);
    std::string cache_path = cache_params.cache_path;
    if (cache_path.empty()) throw FileCacheException("No cache directory specified");
    std::string cache_job_dir_path = cache_params.cache_job_dir_path;
    if(cache_job_dir_path.empty()) throw FileCacheException("No per-job directory specified");
    std::string cache_link_path = cache_params.cache_link_path;
    
    // tidy up paths - take off any trailing slashes
    if (cache_path.rfind("/") == cache_path.length()-1) cache_path = cache_path.substr(0, cache_path.length()-1);
    if (cache_job_dir_path.rfind("/") == cache_job_dir_path.length()-1) cache_job_dir_path = cache_job_dir_path.substr(0, cache_job_dir_path.length()-1);
    if (cache_link_path.rfind("/") == cache_link_path.length()-1) cache_link_path = cache_link_path.substr(0, cache_link_path.length()-1);
  
    // create cache dir
    if (!_cacheMkDir(cache_path, true)) throw FileCacheException("Cannot create directory "+cache_path+" for cache");
  
    // create dir for per-job hard links, making the final dir readable only by the job user
    if (cache_job_dir_path.substr(0, cache_path.length()+1) == cache_path+"/") throw FileCacheException("Cannot have per job link directory in the main cache directory");
    std::string hard_link_path = cache_job_dir_path + "/" + _id;
    if (!_cacheMkDir(hard_link_path, true)) throw FileCacheException("Cannot create directory "+hard_link_path+" for per-job hard links");
    if (chown(hard_link_path.c_str(), _uid, _gid) != 0) throw FileCacheException("Cannot change owner of "+cache_path);
    if (chmod(hard_link_path.c_str(), S_IRWXU) != 0) throw FileCacheException("Cannot change permissions of "+cache_path+" to 0700");

    // add this cache to our list
    cache_params.cache_path = cache_path;
    cache_params.cache_job_dir_path = cache_job_dir_path;
    cache_params.cache_link_path = cache_link_path;
    _caches.push_back(cache_params);
  }
  
  // our hostname and pid
  struct utsname buf;
  if (uname(&buf) != 0) throw FileCacheException("Cannot determine hostname from uname()");
  _hostname = buf.nodename;
  int pid_i = getpid();
  std::stringstream ss;
  ss << pid_i;
  ss >> _pid;
}

    
FileCache::FileCache(const FileCache& cache) {
  _caches = cache._caches;
  _cache_path = cache._cache_path;
  _cache_job_dir_path = cache._cache_job_dir_path;
  _cache_link_path = cache._cache_link_path;
  _id = cache._id;
  _uid = cache._uid;
  _gid = cache._gid;
  
  // our hostname and pid
  struct utsname buf;
  if (uname(&buf) != 0) throw FileCacheException("Cannot determine hostname from uname()");
  _hostname = buf.nodename;
  int pid_i = getpid();
  std::stringstream ss;
  ss << pid_i;
  ss >> _pid;
  
}

FileCache::~FileCache(void) {
}

bool FileCache::start(std::string url, bool &available, bool &is_locked) {
	
  available = false;
  is_locked = false;
	std::string filename = file(url);
  std::string lock_file = _getLockFileName(url);

  // create directory structure if required, only readable by GM user
  if (!_cacheMkDir(lock_file.substr(0, lock_file.rfind("/")), false)) return false;
  
  int lock_timeout = 86400; // one day timeout on lock TODO: make configurable?
  
  // locking mechanism:
  // - check if lock is there
  // - if not, create tmp file and check again
  // - if lock is still not there copy tmp file to cache lock file
  // - check pid inside lock file matches ours
  
  struct stat fileStat;
  int err = stat( lock_file.c_str(), &fileStat ); 
  if (0 != err) {
    if (errno == EACCES) {
      odlog(ERROR)<<"EACCES Error opening lock file"<<lock_file<<": "<<strerror(errno)<<std::endl;
      return false;
    }
    else if (errno != ENOENT) {
      // some other error occurred opening the lock file
      odlog(ERROR)<<"Error opening lock file "<<lock_file<<" in initial check: "<<strerror(errno)<<std::endl;
      return false;
    }
    // lock does not exist - create tmp file
    // ugly char manipulation to get mkstemp() to work...
    char tmpfile[256];
    tmpfile[0] = '\0';
    strcat(tmpfile, lock_file.c_str());
    strcat(tmpfile, ".XXXXXX");
    int h = mkstemp(tmpfile);
    if (h == -1) {
      odlog(ERROR)<<"Error creating file "<<tmpfile<<" with mkstemp(): "<<strerror(errno)<<std::endl;
      return false;
    }
    // write pid@hostname to the lock file
    char buf[_pid.length()+_hostname.length()+2];
    sprintf(buf, "%s@%s", _pid.c_str(), _hostname.c_str());
    if (write(h, &buf, strlen(buf)) == -1) {
      odlog(ERROR)<<"Error writing to tmp lock file "<<tmpfile<<": "<<strerror(errno)<<std::endl;
      close(h);
      // not much we can do if this doesn't work, but it is only a tmp file
      remove(tmpfile);
      return false;
    }
    if (close(h) != 0) {
      // not critical as file will be removed after we are done
      odlog(WARNING)<<"Warning: closing tmp lock file "<<tmpfile<<" failed"<<std::endl;
    }
    // check again if lock exists, in case creating the tmp file took some time
    err = stat( lock_file.c_str(), &fileStat ); 
    if (0 != err) {
      if (errno == ENOENT) {
        // ok, we can create lock
        if (rename(tmpfile, lock_file.c_str()) != 0) {
          odlog(ERROR)<<"Error renaming tmp file "<<tmpfile<<" to lock file "<<lock_file<<": "<<strerror(errno)<<std::endl;
          remove(tmpfile);
          return false;
        }
        // check it's really there
        err = stat( lock_file.c_str(), &fileStat ); 
        if (0 != err) {
          odlog(ERROR)<<"Error renaming lock file, even though rename() did not return an error"<<std::endl;
          return false;
        }
        // check the pid inside the lock file, just in case...
        if (!_checkLock(url)) {
          is_locked = true;
          return false;
        }
      }
      else if (errno == EACCES) {
        odlog(ERROR)<<"EACCES Error opening lock file"<<lock_file<<": "<<strerror(errno)<<std::endl;
        return false;
      }
      else {
        // some other error occurred opening the lock file
        odlog(ERROR)<<"Error opening lock file we just renamed successfully "<<lock_file<<": "<<strerror(errno)<<std::endl;
        return false;
      }
    }
    else {
      odlog(DEBUG)<<"The file is currently locked with a valid lock"<<std::endl;
      is_locked = true;
      return false;
    }
  }
  else {
    // the lock already exists, check if it has expired
    // look at modification time
    time_t mod_time = fileStat.st_mtime;
    time_t now = time(NULL);
    odlog(DEBUG)<<(now - mod_time)<<" seconds since lock file was created"<<std::endl;
    
    if ((now - mod_time) > lock_timeout) {
      odlog(DEBUG)<<"Timeout has expired, will remove lock file"<<std::endl;
      // TODO: kill the process holding the lock, only if we know it was the original
      // process which created it
      if (remove(lock_file.c_str()) != 0 && errno != ENOENT) {
        odlog(ERROR)<<"Failed to unlock file "<<lock_file<<": "<<strerror(errno)<<std::endl;
        return false;
      }
      // lock has expired and has been removed. Try to remove cache file and call start() again
      if (remove(filename.c_str()) != 0 && errno != ENOENT) {
        odlog(ERROR)<<"Error removing cache file "<<file(url)<<": "<<strerror(errno)<<std::endl;
        return false;
      }
      return start(url, available, is_locked);
    }
    
    // lock is still valid, check if we own it
    FILE * pFile;
    char lock_info [100]; // should be long enough for a pid + hostname
    pFile = fopen ((char*)lock_file.c_str(), "r");
    if (pFile == NULL) {
      // lock could have been released by another process, so call start again
      if (errno == ENOENT) {
        odlog(DEBUG)<<"Lock that recently existed has been deleted by another process, calling start() again"<<std::endl;
        return start(url, available, is_locked);
      }
      odlog(ERROR)<<"Error opening valid and existing lock file "<<lock_file<<": "<<strerror(errno)<<std::endl;
      return false;
    }
    fgets (lock_info, 100, pFile);
    fclose (pFile);
    
    std::string lock_info_s(lock_info);
    std::string::size_type index = lock_info_s.find("@", 0);
    if (index == std::string::npos) {
      odlog(ERROR)<<"Error with formatting in lock file "<<lock_file<<": "<<lock_info_s<<std::endl;
      return false;
    }
    
    if (lock_info_s.substr(index+1) != _hostname) {
      odlog(DEBUG)<<"Lock is owned by a different host"<<std::endl;
      // TODO: here do ssh login and check
      is_locked = true;
      return false;
    }
    std::string lock_pid = lock_info_s.substr(0, index);
    if (lock_pid == _pid) {
      // safer to wait until lock expires than use cached file or re-download
      odlog(WARNING)<<"Warning: This process already owns the lock"<<std::endl;
    }
    else {
      // check if the pid owning the lock is still running - if not we can claim the lock
      // this is not really portable... but no other way to do it
      std::string procdir("/proc/");
      procdir = procdir.append(lock_pid);
      if (stat(procdir.c_str(), &fileStat) != 0 && errno == ENOENT) {
        odlog(DEBUG)<<"The process owning the lock is no longer running, will remove lock"<<std::endl;
        if (remove(lock_file.c_str()) != 0) {
          odlog(ERROR)<<"Failed to unlock file "<<lock_file<<": "<<strerror(errno)<<std::endl;
          return false;
        }
        // lock has been removed. try to delete cache file and call start() again
        if (remove(filename.c_str()) != 0 && errno != ENOENT) {
          odlog(ERROR)<<"Error removing cache file "<<file(url)<<": "<<strerror(errno)<<std::endl;
          return false;
        }
        return start(url, available, is_locked);
      }
    }
    
    odlog(DEBUG)<<"The file is currently locked with a valid lock"<<std::endl;
    is_locked = true;
    return false;
  }

  // if we get to here we have acquired the lock

  // create the meta file to store the URL, if it does not exist
  std::string meta_file = _getMetaFileName(url);
  err = stat( meta_file.c_str(), &fileStat ); 
  if (0 == err) {
    // check URL inside file for possible hash collisions
    FILE * pFile;
    char mystring [1024]; // should be long enough for a pid or url...
    pFile = fopen ((char*)_getMetaFileName(url).c_str(), "r");
    if (pFile == NULL) {
      odlog(ERROR)<<"Error opening meta file "<<_getMetaFileName(url)<<": "<<strerror(errno)<<std::endl;
      remove(lock_file.c_str());
      return false;
    }
    fgets (mystring, sizeof(mystring), pFile);
    fclose (pFile);
    
    std::string meta_str(mystring);
    std::string::size_type space_pos = meta_str.find(' ', 0);
    if (space_pos == std::string::npos) space_pos = meta_str.length(); 
    if (meta_str.substr(0, space_pos) != url) {
      odlog(ERROR)<<"Error: File "<<url<<" is already cached at "<<filename<<" under a different URL: "<<meta_str.substr(0, space_pos)<<" - this file will not be cached"<<std::endl;
      remove(lock_file.c_str());
      return false;
    }
  }
  else if (errno == ENOENT) {
    // create new file
    FILE * pFile;
    pFile = fopen ((char*)meta_file.c_str(), "w");
    if (pFile == NULL) {
      odlog(ERROR)<<"Failed to create info file: "<<meta_file<<": "<<strerror(errno)<<std::endl;
      remove(lock_file.c_str());
      return false;
    }
    fputs ((char*)url.c_str(), pFile);
    fclose (pFile);
    // make read/writeable only by GM user
    chmod(meta_file.c_str(), S_IRUSR | S_IWUSR);
  }
  else {
    odlog(ERROR)<<"Error looking up attributes of meta file "<<meta_file<<": "<<strerror(errno)<<std::endl;
    remove(lock_file.c_str());
    return false;
  }
  // now check if the cache file is there already
  err = stat( filename.c_str(), &fileStat );
  if (0 == err)  available = true;
  else if (errno != ENOENT) {
    // this is ok, we will download again
    odlog(WARNING)<<"Warning: error looking up attributes of cached file: "<<strerror(errno)<<std::endl;
  }
  return true;
}

bool FileCache::stop(std::string url) {
  
  // check the lock is ok to delete
  if (!_checkLock(url)) return false;

  // delete the lock
  if (remove(_getLockFileName(url).c_str()) != 0) {
    odlog(ERROR)<<"Failed to unlock file with lock "<<_getLockFileName(url)<<": "<<strerror(errno)<<std::endl;
    return false;
  }
  return true;
}

bool FileCache::stopAndDelete(std::string url) {
  
  // check the lock is ok to delete, and if so, remove the file and the
  // associated lock
  if (!_checkLock(url)) return false;
  
  // delete the cache file
  if (remove(file(url).c_str()) != 0 && errno != ENOENT) {
    odlog(ERROR)<<"Error removing cache file "<<file(url)<<": "<<strerror(errno)<<std::endl;
    return false;
  }
  
  // delete the meta file - not critical so don't fail on error
  if (remove(_getMetaFileName(url).c_str()) != 0) {
    odlog(ERROR)<<"Failed to unlock file with lock "<<_getLockFileName(url)<<": "<<strerror(errno)<<std::endl;
  }
  
  // delete the lock
  if (remove(_getLockFileName(url).c_str()) != 0) {
    odlog(ERROR)<<"Failed to unlock file with lock "<<_getLockFileName(url)<<": "<<strerror(errno)<<std::endl;
    return false;
  }
  return true;
}

std::string FileCache::file(std::string url) {
  
  // get the hash of the url
  std::string hash = FileCacheHash::getHash(url);
  
  int index = 0;
  for(int level = 0; level < CACHE_DIR_LEVELS; level ++) {
    hash.insert(index + CACHE_DIR_LENGTH, "/");
    // go to next slash position, add one since we just inserted a slash
    index += CACHE_DIR_LENGTH + 1;
  }
  return _caches.at(_chooseCache(hash)).cache_path+"/"+hash;
}

bool FileCache::link_file(std::string link_path, std::string url) {

  // choose cache
  struct CacheParameters cache_params = _caches.at(_chooseCache(FileCacheHash::getHash(url)));
  
  // if _cache_link_path is '.' then copy instead, bypassing the hard-link
  if (cache_params.cache_link_path == ".") return copy_file(link_path, url);
   
  std::string hard_link_path = cache_params.cache_job_dir_path + "/" + _id;
  std::string filename = link_path.substr(link_path.rfind("/")+1);
  std::string hard_link_file = hard_link_path + "/" + filename;
  std::string session_dir = link_path.substr(0, link_path.rfind("/"));
  
  // check the original file exists
  struct stat fileStat;
  if (stat(file(url).c_str(), &fileStat) != 0) {
    if (errno == ENOENT) odlog(ERROR)<<"Error: Cache file "<<file(url)<<" does not exist"<<std::endl;
    else odlog(ERROR)<<"Error accessing cache file "<<file(url)<<": "<<strerror(errno)<<std::endl;
    return false;
  }
  
  // make the hard link
  if (link(file(url).c_str(), hard_link_file.c_str()) != 0) {
    odlog(ERROR)<<"Failed to create hard link from "<<hard_link_file<<" to "<<file(url)<<": "<<strerror(errno)<<std::endl;
    return false;
  }
  // ensure the hard link is readable by all and owned by root (or GM user)
  // (to make cache file immutable but readable by all)
  if (chown(hard_link_file.c_str(), getuid(), getgid()) != 0) {
    odlog(ERROR)<<"Failed to change owner of hard link to "<<getuid()<<": "<<strerror(errno)<<std::endl;
    return false;
  }
  if (chmod(hard_link_file.c_str(), S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH) != 0) {
    odlog(ERROR)<<"Failed to change permissions of hard link to 0644: "<<strerror(errno)<<std::endl;
    return false;
  }
  
  // make necessary dirs for the soft link
  // this probably should have already been done... somewhere...
  if (!_cacheMkDir(session_dir, true)) return false;
  if (chown(session_dir.c_str(), _uid, _gid) != 0) {
    odlog(ERROR)<<"Failed to change owner of session dir to "<<_uid<<": "<<strerror(errno)<<std::endl;
    return false;
  }
  if (chmod(session_dir.c_str(), S_IRWXU) != 0) {
    odlog(ERROR)<<"Failed to change permissions of session dir to 0700: "<<strerror(errno)<<std::endl;
    return false;
  }
  
  // make the soft link, changing the target if cache_link_path is defined
  if (!_cache_link_path.empty()) {
    hard_link_file.replace(0, cache_params.cache_job_dir_path.length(), _cache_link_path, 0, cache_params.cache_link_path.length());
  }
  if (symlink(hard_link_file.c_str(), link_path.c_str()) != 0) {
    odlog(ERROR)<<"Failed to create soft link: "<<strerror(errno)<<std::endl;
    return false;
  }
  
  // change the owner of the soft link to the job user
  if (lchown(link_path.c_str(), _uid, _gid) != 0) {
     odlog(ERROR)<<"Failed to change owner of session dir to "<<_uid<<": "<<strerror(errno)<<std::endl;
     return false;
  }
  return true;
}

bool FileCache::copy_file(std::string dest_path, std::string url) {
  
  // check the original file exists
  std::string cache_file = file(url);
  struct stat fileStat;
  if (stat(cache_file.c_str(), &fileStat) != 0) {
    if (errno == ENOENT) odlog(ERROR)<<"Error: Cache file "<<cache_file<<" does not exist"<<std::endl;
    else odlog(ERROR)<<"Error accessing cache file "<<cache_file<<": "<<strerror(errno)<<std::endl;
    return false;
  }
  
  // make necessary dirs for the copy
  // this probably should have already been done... somewhere...
  std::string dest_dir = dest_path.substr(0, dest_path.rfind("/"));
  if (!_cacheMkDir(dest_dir, true)) return false;
  if (chown(dest_dir.c_str(), _uid, _gid) != 0) {
    odlog(ERROR)<<"Failed to change owner of destination dir to "<<_uid<<": "<<strerror(errno)<<std::endl;
    return false;
  }
  if (chmod(dest_dir.c_str(), S_IRWXU) != 0) {
    odlog(ERROR)<<"Failed to change permissions of session dir to 0700: "<<strerror(errno)<<std::endl;
    return false;
  }

  // do the copy - taken directly from old datacache.cc
  char buf[65536];
  int fdest = open(dest_path.c_str(), O_WRONLY | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
  if(fdest == -1) {
    odlog(ERROR)<<"Failed to create file "<<dest_path<<" for writing: "<<strerror(errno)<<std::endl;
    return false;
  };
  fchown(fdest, _uid, _gid);
  
  int fsource = open(cache_file.c_str(), O_RDONLY);
  if(fsource == -1) {
    close(fdest);
    odlog(ERROR)<<"Failed to open file "<<cache_file<<" for reading: "<<strerror(errno)<<std::endl;
    return false;
  };
  
  // source and dest opened ok - copy in chunks
  for(;;) {
    ssize_t lin = read(fsource, buf, sizeof(buf));
    if(lin == -1) {
      close(fdest); close(fsource);
      odlog(ERROR)<<"Failed to read file "<<cache_file<<": "<<strerror(errno)<<std::endl;
      return false;
    };
    if(lin == 0) break; // eof
    
    for(ssize_t lout = 0; lout < lin;) {
      ssize_t lwritten = write(fdest, buf+lout, lin-lout);
      if(lwritten == -1) {
        close(fdest); close(fsource);
        odlog(ERROR)<<"Failed to write file "<<dest_path<<": "<<strerror(errno)<<std::endl;
        return false;
      };
      lout += lwritten;
    };
  };
  close(fdest); close(fsource);
  return true;
}

bool FileCache::release() {
  
  // go through all caches and remove per-job dirs for our job id
  for (int i = 0; i < (int)_caches.size(); i++) {
    
    std::string job_dir = _caches.at(i).cache_job_dir_path + "/" + _id;
    
    // check if job dir exists
    DIR * dirp = opendir(job_dir.c_str());
    if ( dirp == NULL) {
      if (errno == ENOENT) continue;
      odlog(ERROR)<<"Error opening per-job dir "<<job_dir<<": "<<strerror(errno)<<std::endl;
      return false;
    }
    
    // list all files in the dir and delete them
    struct dirent *dp;
    errno = 0;
    while ((dp = readdir(dirp)))  {
      if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0) continue;
      std::string to_delete = job_dir + "/" + dp->d_name; 
      odlog(DEBUG)<<"Removing "<<to_delete<<std::endl;
      if (remove(to_delete.c_str()) != 0) {
        odlog(ERROR)<<"Error: failed to remove hard link "<<to_delete<<": "<<strerror(errno)<<std::endl;
        closedir(dirp);
        return false;
      }
    }
    closedir(dirp);
    
    if (errno != 0) {
      odlog(ERROR)<<"Error listing dir "<<job_dir<<": "<<strerror(errno)<<std::endl;
      return false;
    }
    
    // remove now-empty dir
    odlog(DEBUG)<<"Removing "<<job_dir<<std::endl;
    if (rmdir(job_dir.c_str()) != 0) {
      odlog(ERROR)<<"Error: failed to remove cache per-job dir "<<job_dir<<": "<<strerror(errno)<<std::endl;
      return false;
    }
  }
  return true;
}

bool FileCache::created_available(std::string url) {
  
  // check the cache file exists - if so we can get the creation date
  std::string cache_file = file(url);
  struct stat fileStat;
  return (stat(cache_file.c_str(), &fileStat) == 0) ? true : false;
}

time_t FileCache::created(std::string url) {
  
  // check the cache file exists
  std::string cache_file = file(url);
  struct stat fileStat;
  if (stat(cache_file.c_str(), &fileStat) != 0) {
    if (errno == ENOENT) odlog(ERROR)<<"Error: Cache file "<<cache_file<<" does not exist"<<std::endl;
    else odlog(ERROR)<<"Error accessing cache file "<<cache_file<<": "<<strerror(errno)<<std::endl;
    return 0;
  }
  
  time_t ctime = fileStat.st_ctime;
  if (ctime <= 0) return 0;
  return ctime;
}

bool FileCache::validtill_available(std::string url) {
  return ((int)validtill(url) == 0) ? false : true;
}

time_t FileCache::validtill(std::string url) {
  
  // open meta file and pick out expiry time if it exists
  
  FILE * pFile;
  char mystring [1024]; // should be long enough for a pid or url...
  pFile = fopen ((char*)_getMetaFileName(url).c_str(), "r");
  if (pFile == NULL) {
    odlog(ERROR)<<"Error opening meta file "<<_getMetaFileName(url)<<": "<<strerror(errno)<<std::endl;
    return 0;
  }
  fgets (mystring, sizeof(mystring), pFile);
  fclose (pFile);
  
  std::string meta_str(mystring);
  // if the file contains only the url, we don't have an expiry time
  if (meta_str == url) return 0;

  // check sensible formatting - should be like "rls://rls1.ndgf.org/file1 1234567890"
  if (meta_str.substr(0, url.length()+1) != url+" ") {
    odlog(ERROR)<<"Error: mismatching url in file "<<_getMetaFileName(url)<<": "<<meta_str<<" Expected "<<url<<std::endl;
    return 0;
  }
  if (meta_str.length() != url.length() + 11) {
    odlog(ERROR)<<"Error: bad format in file "<<_getMetaFileName(url)<<": "<<meta_str<<std::endl;
    return 0;
  }
  if (meta_str.substr(url.length(), 1) != " ") {
    odlog(ERROR)<<"Error: bad separator in file "<<_getMetaFileName(url)<<": "<<meta_str<<std::endl;
    return 0;    
  }
  if (meta_str.substr(url.length() + 1).length() != 10) {
    odlog(ERROR)<<"Error: bad value of expiry time in "<<_getMetaFileName(url)<<": "<<meta_str<<std::endl;
    return 0;
  }
  
  // convert to int
  int exp_time;
  if(EOF == sscanf(meta_str.substr(url.length() + 1).c_str(), "%i", &exp_time) || exp_time < 0) {
    odlog(ERROR)<<"Error with converting time in file "<<_getMetaFileName(url)<<": "<<meta_str<<std::endl;
    return 0;
  }
  return (time_t)exp_time;
}

bool FileCache::validtill_force(std::string url, time_t val) {
  
  std::string meta_file = _getMetaFileName(url);
  FILE * pFile;
  pFile = fopen ((char*)meta_file.c_str(), "w");
  if (pFile == NULL) {
    odlog(ERROR)<<"Error opening meta file "<<meta_file<<": "<<strerror(errno)<<std::endl;
    return false;
  }
  // convert time to string
  std::stringstream out;
  out << val;
  std::string file_data = url + " " + out.str();
  fputs ((char*)file_data.c_str(), pFile);
  fclose (pFile);
  return true;
}

bool FileCache::validtill(std::string url, time_t val) {
  // if no time is defined call force, otherwise return false
  return ((int)validtill(url) == 0) ? validtill_force(url, val) : false;
}

bool FileCache::operator==(const FileCache& a) {
  if (a._caches.size() != _caches.size()) return false;
  for (int i = 0; i < (int)a._caches.size(); i++) {
    if (a._caches.at(i).cache_path != _caches.at(i).cache_path) return false;
    if (a._caches.at(i).cache_job_dir_path != _caches.at(i).cache_job_dir_path) return false;
    if (a._caches.at(i).cache_link_path != _caches.at(i).cache_link_path) return false;
  }
  return (
    a._cache_path == _cache_path &&
    a._cache_job_dir_path == _cache_job_dir_path &&
    a._cache_link_path == _cache_link_path &&
    a._id == _id &&
    a._uid == _uid &&
    a._gid == _gid
  );
}
bool FileCache::_checkLock(std::string url) {
  
  std::string filename = file(url);
  std::string lock_file = _getLockFileName(url);
  
  // check for existence of lock file
  struct stat fileStat;
  int err = stat( lock_file.c_str(), &fileStat ); 
  if (0 != err) {
    if (errno == ENOENT) odlog(ERROR)<<"Error: lock file "<<lock_file<<" doesn't exist"<<std::endl;
    else odlog(ERROR)<<"Error listing lock file "<<lock_file<<": "<<strerror(errno)<<std::endl;
    return false;
  }
  
  // check the lock file's pid and hostname matches ours
  FILE * pFile;
  char lock_info [100]; // should be long enough for a pid + hostname
  pFile = fopen ((char*)lock_file.c_str(), "r");
  if (pFile == NULL) {
    odlog(ERROR)<<"Error opening lock file "<<lock_file<<": "<<strerror(errno)<<std::endl;
    return false;
  }
  fgets (lock_info, 100, pFile);
  fclose (pFile);

  std::string lock_info_s(lock_info);
  std::string::size_type index = lock_info_s.find("@", 0);
  if (index == std::string::npos) {
    odlog(ERROR)<<"Error with formatting in lock file "<<lock_file<<": "<<lock_info_s<<std::endl;
    return false;
  }
  
  if (lock_info_s.substr(index+1) != _hostname) {
    odlog(DEBUG)<<"Lock is owned by a different host"<<std::endl;
    // TODO: here do ssh login and check
    return false;
  }
  if (lock_info_s.substr(0, index) != _pid) {
    odlog(ERROR)<<"Error: Another process owns the lock on file "<<filename<<". Must go back to start()"<<std::endl;
    return false;
  }
  return true;
}

std::string FileCache::_getLockFileName(std::string url) {
  return file(url)+CACHE_LOCK_SUFFIX;
}

std::string FileCache::_getMetaFileName(std::string url) {
  return file(url)+CACHE_META_SUFFIX;
}

bool FileCache::_cacheMkDir(std::string dir, bool all_read) {
  
  struct stat fileStat;
  int err = stat( dir.c_str(), &fileStat ); 
  if (0 != err) {
    odlog(DEBUG)<<"Creating directory "<<dir<<std::endl;
    std::string::size_type slashpos = 0;
    
    // set perms based on all_read
    mode_t perm = S_IRWXU;
    if (all_read) perm |= S_IRGRP | S_IROTH | S_IXGRP | S_IXOTH;
    
    do {
      slashpos = dir.find("/", slashpos+1);
      std::string dirname = dir.substr(0, slashpos);
      // list dir to see if it exists (we can't tell the difference between
      // dir already exists and permission denied)
      struct stat statbuf;
      if (stat(dirname.c_str(), &statbuf) == 0) {
        continue;
      };
      if (mkdir(dirname.c_str(), perm) != 0) {
        if (errno != EEXIST) {
          odlog(ERROR)<<"Error creating required dirs: "<<strerror(errno)<<std::endl;
          return false;
        };
      };
      // chmod to get around GM umask setting
      if (chmod(dirname.c_str(), perm) != 0) {
        odlog(ERROR)<<"Error changing permission of dir "<<dirname<<": "<<strerror(errno)<<std::endl;
        return false;
      };
    } while (slashpos != std::string::npos);
  }
  return true;
}

int FileCache::_chooseCache(std::string hash) {
  // choose which cache to use
  // divide based on the first character of the hash, choosing the cache
  // based on the letter mod number of caches
  // this algorithm limits the number of caches to 16
  
  char firstletter = hash.at(0);
  int cacheno = firstletter % _caches.size();
  return cacheno;
}
